@@ -65,6 +65,7 @@
  *  \brief Osmocom stream socket helpers
  */
 
+#ifdef HAVE_LIBSCTP
 /*
  * Platforms that don't have MSG_NOSIGNAL (which disables SIGPIPE)
  * usually have SO_NOSIGPIPE (set via setsockopt).
@@ -73,6 +74,99 @@
 #define MSG_NOSIGNAL 0
 #endif
 
+/* is any of the bytes from offset .. u8_size in 'u8' non-zero? return offset or -1 if all zero */
+static int byte_nonzero(const uint8_t *u8, unsigned int offset, unsigned int u8_size)
+{
+	int j;
+
+	for (j = offset; j < u8_size; j++) {
+		if (u8[j] != 0)
+			return j;
+	}
+
+	return -1;
+}
+
+static int sctp_sockopt_event_subscribe_size = 0;
+
+static int determine_sctp_sockopt_event_subscribe_size(void)
+{
+	uint8_t buf[256];
+	socklen_t buf_len = sizeof(buf);
+	int sd, rc;
+
+	/* only do this once */
+	if (sctp_sockopt_event_subscribe_size != 0)
+		return 0;
+
+	sd = socket(AF_INET, SOCK_STREAM, IPPROTO_SCTP);
+	if (sd < 0)
+		return sd;
+
+	rc = getsockopt(sd, IPPROTO_SCTP, SCTP_EVENTS, buf, &buf_len);
+	close(sd);
+	if (rc < 0)
+		return rc;
+
+	sctp_sockopt_event_subscribe_size = buf_len;
+
+	LOGP(DLINP, LOGL_INFO, "sizes of 'struct sctp_event_subscribe': compile-time %zu, kernel: %u\n",
+		sizeof(struct sctp_event_subscribe), sctp_sockopt_event_subscribe_size);
+	return 0;
+}
+
+/* Attempt to work around Linux kernel ABI breakage
+ *
+ * The Linux kernel ABI for the SCTP_EVENTS socket option has been broken repeatedly.
+ *  - until commit 35ea82d611da59f8bea44a37996b3b11bb1d3fd7 ( kernel < 4.11), the size is 10 bytes
+ *  - in 4.11 it is 11 bytes
+ *  - in 4.12 .. 5.4 it is 13 bytes
+ *  - in kernels >= 5.5 it is 14 bytes
+ *
+ * This wouldn't be a problem if the kernel didn't have a "stupid" assumption that the structure
+ * size passed by userspace will match 1:1 the length of the structure at kernel compile time. In
+ * an ideal world, it would just use the known first bytes and assume the remainder is all zero.
+ * But as it doesn't do that, let's try to work around this */
+static int sctp_setsockopt_events_linux_workaround(int fd, const struct sctp_event_subscribe *event)
+{
+
+	const unsigned int compiletime_size = sizeof(*event);
+	int rc;
+
+	if (determine_sctp_sockopt_event_subscribe_size() < 0) {
+		LOGP(DLINP, LOGL_ERROR, "Cannot determine SCTP_EVENTS socket option size\n");
+		return -1;
+	}
+
+	if (compiletime_size == sctp_sockopt_event_subscribe_size) {
+		/* no kernel workaround needed */
+		return setsockopt(fd, IPPROTO_SCTP, SCTP_EVENTS, event, compiletime_size);
+	} else if (compiletime_size < sctp_sockopt_event_subscribe_size) {
+		/* we are using an older userspace with a more modern kernel and hence need
+		 * to pad the data */
+		uint8_t buf[sctp_sockopt_event_subscribe_size];
+
+		memcpy(buf, event, compiletime_size);
+		memset(buf + sizeof(*event), 0, sctp_sockopt_event_subscribe_size - compiletime_size);
+		return setsockopt(fd, IPPROTO_SCTP, SCTP_EVENTS, buf, sctp_sockopt_event_subscribe_size);
+	} else /* if (compiletime_size > sctp_sockopt_event_subscribe_size) */ {
+		/* we are using a newer userspace with an older kernel and hence need to truncate
+		 * the data - but only if the caller didn't try to enable any of the events of the
+		 * truncated portion */
+		rc = byte_nonzero((const uint8_t *)event, sctp_sockopt_event_subscribe_size,
+				  compiletime_size);
+		if (rc >= 0) {
+			LOGP(DLINP, LOGL_ERROR, "Kernel only supports sctp_event_subscribe of %u bytes, "
+				"but caller tried to enable more modern event at offset %u\n",
+				sctp_sockopt_event_subscribe_size, rc);
+			return -1;
+		}
+
+		return setsockopt(fd, IPPROTO_SCTP, SCTP_EVENTS, event, sctp_sockopt_event_subscribe_size);
+	}
+}
+#endif // HAVE_LIBSCTP
+
 static int sctp_sock_activate_events(int fd)
 {
 #ifdef HAVE_LIBSCTP
@@ -89,12 +183,10 @@ static int sctp_sock_activate_events(int fd)
 	event.sctp_shutdown_event = 1;
 	/* IMPORTANT: Do NOT enable sender_dry_event here, see
 	 * https://bugzilla.redhat.com/show_bug.cgi?id=1442784 */
-	rc = setsockopt(fd, IPPROTO_SCTP, SCTP_EVENTS,
-			&event, sizeof(event));
 
+	rc = sctp_setsockopt_events_linux_workaround(fd, &event);
 	if (rc < 0)
-		LOGP(DLINP, LOGL_ERROR, "couldn't activate SCTP events "
-		     "on FD %u\n", fd);
+		LOGP(DLINP, LOGL_ERROR, "couldn't activate SCTP events on FD %u\n", fd);
 	return rc;
 #else
 	return -1;
@@ -106,9 +198,11 @@ static int setsockopt_nodelay(int fd, int proto, int on)
 	int rc;
 
 	switch (proto) {
+#ifdef HAVE_LIBSCTP
 	case IPPROTO_SCTP:
 		rc = setsockopt(fd, IPPROTO_SCTP, SCTP_NODELAY, &on, sizeof(on));
 		break;
+#endif
 	case IPPROTO_TCP:
 		rc = setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(on));
 		break;
@@ -127,32 +221,40 @@ static int setsockopt_nodelay(int fd, int proto, int on)
  */
 
 enum osmo_stream_cli_state {
-        STREAM_CLI_STATE_NONE         = 0,
-        STREAM_CLI_STATE_CONNECTING   = 1,
-        STREAM_CLI_STATE_CONNECTED    = 2,
-        STREAM_CLI_STATE_MAX
+	STREAM_CLI_STATE_CLOSED,	 /* No fd associated, no timer active */
+	STREAM_CLI_STATE_WAIT_RECONNECT, /* No fd associated, has timer active to try to connect again */
+	STREAM_CLI_STATE_CONNECTING,	 /* Fd associated, but connection not yet confirmed by peer or lower layers */
+	STREAM_CLI_STATE_CONNECTED,	 /* Fd associated and connection is established */
+	STREAM_CLI_STATE_MAX
 };
 
 static const struct value_string stream_cli_state_names[] = {
-	{ STREAM_CLI_STATE_NONE,       "      NONE" },
-	{ STREAM_CLI_STATE_CONNECTING, "CONNECTING" },
-	{ STREAM_CLI_STATE_CONNECTED,  " CONNECTED" },
+	{ STREAM_CLI_STATE_CLOSED,	 "CLOSED" },
+	{ STREAM_CLI_STATE_WAIT_RECONNECT, "WAIT_RECONNECT" },
+	{ STREAM_CLI_STATE_CONNECTING,     "CONNECTING" },
+	{ STREAM_CLI_STATE_CONNECTED,      "CONNECTED" },
 	{ 0, NULL }
 };
 
 #define OSMO_STREAM_CLI_F_RECONF	(1 << 0)
 #define OSMO_STREAM_CLI_F_NODELAY	(1 << 1)
 
+#ifdef HAVE_LIBSCTP
+#define OSMO_STREAM_MAX_ADDRS OSMO_SOCK_MAX_ADDRS
+#else
+#define OSMO_STREAM_MAX_ADDRS 1
+#endif
+
 struct osmo_stream_cli {
 	struct osmo_fd			ofd;
 	struct llist_head		tx_queue;
 	struct osmo_timer_list		timer;
 	enum osmo_stream_cli_state	state;
-	char				*addr[OSMO_SOCK_MAX_ADDRS];
-	uint8_t                         addrcnt;
+	char				*addr[OSMO_STREAM_MAX_ADDRS];
+	uint8_t			 	addrcnt;
 	uint16_t			port;
-	char				*local_addr[OSMO_SOCK_MAX_ADDRS];
-	uint8_t                         local_addrcnt;
+	char				*local_addr[OSMO_STREAM_MAX_ADDRS];
+	uint8_t			 	local_addrcnt;
 	uint16_t			local_port;
 	uint16_t			proto;
 	int (*connect_cb)(struct osmo_stream_cli *srv);
@@ -179,14 +281,15 @@ void osmo_stream_cli_reconnect(struct osmo_stream_cli *cli)
 		return;
 	}
 
+	cli->state = STREAM_CLI_STATE_WAIT_RECONNECT;
 	LOGSCLI(cli, LOGL_INFO, "retrying in %d seconds...\n",
 		cli->reconnect_timeout);
 	osmo_timer_schedule(&cli->timer, cli->reconnect_timeout, 0);
-	cli->state = STREAM_CLI_STATE_CONNECTING;
 }
 
 /*! \brief Check if Osmocom Stream Client is in connected state
  *  \param[in] cli Osmocom Stream Client
+ *  \return true if connected, false otherwise
  */
 bool osmo_stream_cli_is_connected(struct osmo_stream_cli *cli)
 {
@@ -211,7 +314,7 @@ void osmo_stream_cli_close(struct osmo_stream_cli *cli)
 			cli->disconnect_cb(cli);
 	}
 
-	cli->state = STREAM_CLI_STATE_NONE;
+	cli->state = STREAM_CLI_STATE_CLOSED;
 }
 
 static void osmo_stream_cli_read(struct osmo_stream_cli *cli)
@@ -231,21 +334,21 @@ static int osmo_stream_cli_write(struct osmo_stream_cli *cli)
 	struct llist_head *lh;
 	int ret;
 
-	LOGSCLI(cli, LOGL_DEBUG, "sending data\n");
-
 	if (llist_empty(&cli->tx_queue)) {
-		cli->ofd.when &= ~BSC_FD_WRITE;
+		osmo_fd_write_disable(&cli->ofd);
 		return 0;
 	}
 	lh = cli->tx_queue.next;
 	llist_del(lh);
 	msg = llist_entry(lh, struct msgb, list);
 
-	if (cli->state == STREAM_CLI_STATE_CONNECTING) {
+	if (!osmo_stream_cli_is_connected(cli)) {
 		LOGSCLI(cli, LOGL_ERROR, "not connected, dropping data!\n");
 		return 0;
 	}
 
+	LOGSCLI(cli, LOGL_DEBUG, "sending %u bytes of data\n", msgb_length(msg));
+
 	switch (cli->proto) {
 #ifdef HAVE_LIBSCTP
 	case IPPROTO_SCTP:
@@ -258,14 +361,14 @@ static int osmo_stream_cli_write(struct osmo_stream_cli *cli)
 #endif
 	case IPPROTO_TCP:
 	default:
-		ret = send(cli->ofd.fd, msg->data, msg->len, 0);
+		ret = send(cli->ofd.fd, msg->data, msgb_length(msg), 0);
 		break;
 	}
 	if (ret < 0) {
 		if (errno == EPIPE || errno == ENOTCONN) {
 			osmo_stream_cli_reconnect(cli);
 		}
-		LOGSCLI(cli, LOGL_ERROR, "error to send\n");
+		LOGSCLI(cli, LOGL_ERROR, "error %d to send\n", ret);
 	}
 	msgb_free(msg);
 	return 0;
@@ -288,7 +391,7 @@ static int osmo_stream_cli_fd_cb(struct osmo_fd *ofd, unsigned int what)
 		/* If messages got enqueued while 'connecting', keep WRITE flag
 		   up to dispatch them upon next main loop step */
 		if (llist_empty(&cli->tx_queue))
-			cli->ofd.when &= ~BSC_FD_WRITE;
+			osmo_fd_write_disable(&cli->ofd);
 
 		LOGSCLI(cli, LOGL_DEBUG, "connection done.\n");
 		cli->state = STREAM_CLI_STATE_CONNECTED;
@@ -306,19 +409,20 @@ static int osmo_stream_cli_fd_cb(struct osmo_fd *ofd, unsigned int what)
 			cli->connect_cb(cli);
 		break;
 	case STREAM_CLI_STATE_CONNECTED:
-		if (what & BSC_FD_READ) {
+		if (what & OSMO_FD_READ) {
 			LOGSCLI(cli, LOGL_DEBUG, "connected read\n");
 			osmo_stream_cli_read(cli);
 		}
-		if (what & BSC_FD_WRITE) {
+		if (what & OSMO_FD_WRITE) {
 			LOGSCLI(cli, LOGL_DEBUG, "connected write\n");
 			osmo_stream_cli_write(cli);
 		}
 		break;
 	default:
-		break;
+		/* Only CONNECTING and CONNECTED states are expected, since they are the only states where FD exists: */
+		osmo_panic("osmo_stream_cli_fd_cb called with unexpected state %d\n", cli->state);
 	}
-        return 0;
+	return 0;
 }
 
 static void cli_timer_cb(void *data);
@@ -326,7 +430,9 @@ static void cli_timer_cb(void *data);
 /*! \brief Create an Osmocom stream client
  *  \param[in] ctx talloc context from which to allocate memory
  *  This function allocates a new \ref osmo_stream_cli and initializes
- *  it with default values (5s reconnect timer, TCP protocol) */
+ *  it with default values (5s reconnect timer, TCP protocol)
+ *  \return allocated stream client, or NULL in case of error
+ */
 struct osmo_stream_cli *osmo_stream_cli_create(void *ctx)
 {
 	struct osmo_stream_cli *cli;
@@ -337,11 +443,10 @@ struct osmo_stream_cli *osmo_stream_cli_create(void *ctx)
 
 	cli->proto = IPPROTO_TCP;
 	cli->ofd.fd = -1;
-	cli->ofd.when |= BSC_FD_READ | BSC_FD_WRITE;
 	cli->ofd.priv_nr = 0;	/* XXX */
 	cli->ofd.cb = osmo_stream_cli_fd_cb;
 	cli->ofd.data = cli;
-	cli->state = STREAM_CLI_STATE_CONNECTING;
+	cli->state = STREAM_CLI_STATE_CLOSED;
 	osmo_timer_setup(&cli->timer, cli_timer_cb, cli);
 	cli->reconnect_timeout = 5;	/* default is 5 seconds. */
 	INIT_LLIST_HEAD(&cli->tx_queue);
@@ -369,7 +474,7 @@ int osmo_stream_cli_set_addrs(struct osmo_stream_cli *cli, const char **addr, si
 {
 	int i = 0;
 
-	if (addrcnt > OSMO_SOCK_MAX_ADDRS)
+	if (addrcnt > OSMO_STREAM_MAX_ADDRS)
 		return -EINVAL;
 
 	for (; i < addrcnt; i++)
@@ -426,7 +531,7 @@ int osmo_stream_cli_set_local_addrs(struct osmo_stream_cli *cli, const char **ad
 {
 	int i = 0;
 
-	if (addrcnt > OSMO_SOCK_MAX_ADDRS)
+	if (addrcnt > OSMO_STREAM_MAX_ADDRS)
 		return -EINVAL;
 
 	for (; i < addrcnt; i++)
@@ -542,6 +647,7 @@ void osmo_stream_cli_destroy(struct osmo_stream_cli *cli)
  * Open connection of an Osmocom stream client
  *  \param[in] cli Stream Client to connect
  *  \param[in] reconect 1 if we should not automatically reconnect
+ *  \return negative on error, 0 on success
  */
 int osmo_stream_cli_open2(struct osmo_stream_cli *cli, int reconnect)
 {
@@ -554,14 +660,16 @@ int osmo_stream_cli_open2(struct osmo_stream_cli *cli, int reconnect)
 	cli->flags &= ~OSMO_STREAM_CLI_F_RECONF;
 
 	switch (cli->proto) {
+#ifdef HAVE_LIBSCTP
 	case IPPROTO_SCTP:
-		ret = osmo_sock_init2_multiaddr(AF_INET, SOCK_STREAM, cli->proto,
+		ret = osmo_sock_init2_multiaddr(AF_UNSPEC, SOCK_STREAM, cli->proto,
 						(const char **)cli->local_addr, cli->local_addrcnt, cli->local_port,
 						(const char **)cli->addr, cli->addrcnt, cli->port,
 						OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
 		break;
+#endif
 	default:
-		ret = osmo_sock_init2(AF_INET, SOCK_STREAM, cli->proto,
+		ret = osmo_sock_init2(AF_UNSPEC, SOCK_STREAM, cli->proto,
 				      cli->local_addr[0], cli->local_port,
 				      cli->addr[0], cli->port,
 				      OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
@@ -572,7 +680,7 @@ int osmo_stream_cli_open2(struct osmo_stream_cli *cli, int reconnect)
 			osmo_stream_cli_reconnect(cli);
 		return ret;
 	}
-	cli->ofd.fd = ret;
+	osmo_fd_setup(&cli->ofd, ret, OSMO_FD_READ | OSMO_FD_WRITE, cli->ofd.cb, cli->ofd.data, cli->ofd.priv_nr);
 
 	if (cli->flags & OSMO_STREAM_CLI_F_NODELAY) {
 		ret = setsockopt_nodelay(cli->ofd.fd, cli->proto, 1);
@@ -583,6 +691,7 @@ int osmo_stream_cli_open2(struct osmo_stream_cli *cli, int reconnect)
 	if (osmo_fd_register(&cli->ofd) < 0)
 		goto error_close_socket;
 
+	cli->state = STREAM_CLI_STATE_CONNECTING;
 	return 0;
 
 error_close_socket:
@@ -610,7 +719,8 @@ void osmo_stream_cli_set_nodelay(struct osmo_stream_cli *cli, bool nodelay)
 /*! \brief Open connection of an Osmocom stream client
  *  By default the client will automatically reconnect after default timeout.
  *  To disable this, use osmo_stream_cli_set_reconnect_timeout() before calling this function.
- *  \param[in] cli Stream Client to connect */
+ *  \param[in] cli Stream Client to connect
+ *  \return negative on error, 0 on success */
 int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 {
 	int ret;
@@ -623,14 +733,16 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 
 
 	switch (cli->proto) {
+#ifdef HAVE_LIBSCTP
 	case IPPROTO_SCTP:
-		ret = osmo_sock_init2_multiaddr(AF_INET, SOCK_STREAM, cli->proto,
+		ret = osmo_sock_init2_multiaddr(AF_UNSPEC, SOCK_STREAM, cli->proto,
 						(const char **)cli->local_addr, cli->local_addrcnt, cli->local_port,
 						(const char **)cli->addr, cli->addrcnt, cli->port,
 						OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
 		break;
+#endif
 	default:
-		ret = osmo_sock_init2(AF_INET, SOCK_STREAM, cli->proto,
+		ret = osmo_sock_init2(AF_UNSPEC, SOCK_STREAM, cli->proto,
 				      cli->local_addr[0], cli->local_port,
 				      cli->addr[0], cli->port,
 				      OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
@@ -640,7 +752,7 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 		osmo_stream_cli_reconnect(cli);
 		return ret;
 	}
-	cli->ofd.fd = ret;
+	osmo_fd_setup(&cli->ofd, ret, OSMO_FD_READ | OSMO_FD_WRITE, cli->ofd.cb, cli->ofd.data, cli->ofd.priv_nr);
 
 	if (cli->flags & OSMO_STREAM_CLI_F_NODELAY) {
 		ret = setsockopt_nodelay(cli->ofd.fd, cli->proto, 1);
@@ -651,9 +763,11 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 	if (osmo_fd_register(&cli->ofd) < 0)
 		goto error_close_socket;
 
+	cli->state = STREAM_CLI_STATE_CONNECTING;
 	return 0;
 
 error_close_socket:
+	cli->state = STREAM_CLI_STATE_CLOSED;
 	close(cli->ofd.fd);
 	cli->ofd.fd = -1;
 	return -EIO;
@@ -664,15 +778,7 @@ static void cli_timer_cb(void *data)
 	struct osmo_stream_cli *cli = data;
 
 	LOGSCLI(cli, LOGL_DEBUG, "reconnecting.\n");
-
-	switch(cli->state) {
-	case STREAM_CLI_STATE_CONNECTING:
-		cli->ofd.when |= BSC_FD_READ | BSC_FD_WRITE;
-		osmo_stream_cli_open(cli);
-	        break;
-	default:
-		break;
-	}
+	osmo_stream_cli_open(cli);
 }
 
 /*! \brief Enqueue data to be sent via an Osmocom stream client
@@ -681,7 +787,7 @@ static void cli_timer_cb(void *data)
 void osmo_stream_cli_send(struct osmo_stream_cli *cli, struct msgb *msg)
 {
 	msgb_enqueue(&cli->tx_queue, msg);
-	cli->ofd.when |= BSC_FD_WRITE;
+	osmo_fd_write_enable(&cli->ofd);
 }
 
 /*! \brief Receive data via an Osmocom stream client
@@ -717,21 +823,23 @@ int osmo_stream_cli_recv(struct osmo_stream_cli *cli, struct msgb *msg)
 #define OSMO_STREAM_SRV_F_NODELAY	(1 << 1)
 
 struct osmo_stream_srv_link {
-        struct osmo_fd                  ofd;
-        char                            *addr[OSMO_SOCK_MAX_ADDRS];
-        uint8_t                         addrcnt;
-        uint16_t                        port;
-        uint16_t                        proto;
-        int (*accept_cb)(struct osmo_stream_srv_link *srv, int fd);
-        void                            *data;
-	int				flags;
+	struct osmo_fd		ofd;
+	char			*addr[OSMO_STREAM_MAX_ADDRS];
+	uint8_t			addrcnt;
+	uint16_t		port;
+	uint16_t		proto;
+	int (*accept_cb)(struct osmo_stream_srv_link *srv, int fd);
+	void			*data;
+	int			flags;
 };
 
 static int osmo_stream_srv_fd_cb(struct osmo_fd *ofd, unsigned int what)
 {
 	int ret;
 	int sock_fd;
-	struct sockaddr_in sa;
+	char addrstr[128];
+	bool is_ipv6;
+	struct sockaddr_storage sa;
 	socklen_t sa_len = sizeof(sa);
 	struct osmo_stream_srv_link *link = ofd->data;
 
@@ -741,8 +849,13 @@ static int osmo_stream_srv_fd_cb(struct osmo_fd *ofd, unsigned int what)
 			"peer, reason=`%s'\n", strerror(errno));
 		return ret;
 	}
+	is_ipv6 = ((struct sockaddr *)&sa)->sa_family == AF_INET6;
 	LOGP(DLINP, LOGL_DEBUG, "accept()ed new link from %s to port %u\n",
-		inet_ntoa(sa.sin_addr), link->port);
+		inet_ntop(is_ipv6 ? AF_INET6 : AF_INET,
+			  is_ipv6 ? (void*)&(((struct sockaddr_in6 *)&sa)->sin6_addr) :
+				    (void*)&(((struct sockaddr_in *)&sa)->sin_addr),
+			  addrstr, sizeof(addrstr)),
+		link->port);
 	sock_fd = ret;
 
 	if (link->proto == IPPROTO_SCTP) {
@@ -787,10 +900,7 @@ struct osmo_stream_srv_link *osmo_stream_srv_link_create(void *ctx)
 		return NULL;
 
 	link->proto = IPPROTO_TCP;
-	link->ofd.fd = -1;
-	link->ofd.when |= BSC_FD_READ | BSC_FD_WRITE;
-	link->ofd.cb = osmo_stream_srv_fd_cb;
-	link->ofd.data = link;
+	osmo_fd_setup(&link->ofd, -1, OSMO_FD_READ | OSMO_FD_WRITE, osmo_stream_srv_fd_cb, link, 0);
 
 	return link;
 }
@@ -830,7 +940,7 @@ int osmo_stream_srv_link_set_addrs(struct osmo_stream_srv_link *link, const char
 {
 	int i = 0;
 
-	if (addrcnt > OSMO_SOCK_MAX_ADDRS)
+	if (addrcnt > OSMO_STREAM_MAX_ADDRS)
 		return -EINVAL;
 
 	for (; i < addrcnt; i++)
@@ -934,7 +1044,8 @@ void osmo_stream_srv_link_destroy(struct osmo_stream_srv_link *link)
 
 /*! \brief Open the stream server link.  This actually initializes the
  *  underlying socket and binds it to the configured ip/port
- *  \param[in] link Stream Server Link to open */
+ *  \param[in] link Stream Server Link to open
+ *  \return negative on error, 0 on success */
 int osmo_stream_srv_link_open(struct osmo_stream_srv_link *link)
 {
 	int ret;
@@ -950,13 +1061,15 @@ int osmo_stream_srv_link_open(struct osmo_stream_srv_link *link)
 	link->flags &= ~OSMO_STREAM_SRV_F_RECONF;
 
 	switch (link->proto) {
+#ifdef HAVE_LIBSCTP
 	case IPPROTO_SCTP:
-		ret = osmo_sock_init2_multiaddr(AF_INET, SOCK_STREAM, link->proto,
+		ret = osmo_sock_init2_multiaddr(AF_UNSPEC, SOCK_STREAM, link->proto,
 						(const char **)link->addr, link->addrcnt, link->port,
 						NULL, 0, 0, OSMO_SOCK_F_BIND);
 		break;
+#endif
 	default:
-		ret = osmo_sock_init(AF_INET, SOCK_STREAM, link->proto,
+		ret = osmo_sock_init(AF_UNSPEC, SOCK_STREAM, link->proto,
 					link->addr[0], link->port, OSMO_SOCK_F_BIND);
 	}
 	if (ret < 0)
@@ -987,11 +1100,11 @@ void osmo_stream_srv_link_close(struct osmo_stream_srv_link *link)
 
 struct osmo_stream_srv {
 	struct osmo_stream_srv_link	*srv;
-        struct osmo_fd                  ofd;
-        struct llist_head               tx_queue;
-        int (*closed_cb)(struct osmo_stream_srv *peer);
-        int (*cb)(struct osmo_stream_srv *peer);
-        void                            *data;
+	struct osmo_fd			ofd;
+	struct llist_head		tx_queue;
+	int (*closed_cb)(struct osmo_stream_srv *peer);
+	int (*cb)(struct osmo_stream_srv *peer);
+	void				*data;
 	int				flags;
 };
 
@@ -1024,7 +1137,7 @@ static void osmo_stream_srv_write(struct osmo_stream_srv *conn)
 	LOGP(DLINP, LOGL_DEBUG, "sending data\n");
 
 	if (llist_empty(&conn->tx_queue)) {
-		conn->ofd.when &= ~BSC_FD_WRITE;
+		osmo_fd_write_disable(&conn->ofd);
 		return;
 	}
 	lh = conn->tx_queue.next;
@@ -1061,9 +1174,9 @@ static int osmo_stream_srv_cb(struct osmo_fd *ofd, unsigned int what)
 	int rc = 0;
 
 	LOGP(DLINP, LOGL_DEBUG, "connected read/write\n");
-	if (what & BSC_FD_READ)
+	if (what & OSMO_FD_READ)
 		rc = osmo_stream_srv_read(conn);
-	if (rc != -EBADF && (what & BSC_FD_WRITE))
+	if (rc != -EBADF && (what & OSMO_FD_WRITE))
 		osmo_stream_srv_write(conn);
 
 	return rc;
@@ -1088,10 +1201,7 @@ osmo_stream_srv_create(void *ctx, struct osmo_stream_srv_link *link,
 		return NULL;
 	}
 	conn->srv = link;
-	conn->ofd.fd = fd;
-	conn->ofd.data = conn;
-	conn->ofd.cb = osmo_stream_srv_cb;
-	conn->ofd.when = BSC_FD_READ;
+	osmo_fd_setup(&conn->ofd, fd, OSMO_FD_READ, osmo_stream_srv_cb, conn, 0);
 	conn->cb = cb;
 	conn->closed_cb = closed_cb;
 	conn->data = data;
@@ -1177,7 +1287,7 @@ void osmo_stream_srv_send(struct osmo_stream_srv *conn, struct msgb *msg)
 	}
 
 	msgb_enqueue(&conn->tx_queue, msg);
-	conn->ofd.when |= BSC_FD_WRITE;
+	osmo_fd_write_enable(&conn->ofd);
 }
 
 /*! \brief Receive data via Osmocom stream server
